package algo.branch;

import algo.datastructure.*;
import algo.heuristic.Heuristic;
import algo.test.Clustering;

import java.util.*;

import static algo.branch.LowerBound.lb1Count;
import static algo.branch.LowerBound.lb2Count;
import static algo.util.Util.log;

public class Branch {
    private static int branchTreeNodeCounter;

    private static Graph g;
    private static int t;

    static int budget;
    static int resWith;
    static int noResWith;

    private static final Stack<Edge> mergedEdges = new Stack<>();

    private static int initLB = 0;
    private static int heuristicRes = -1;

    //settings
    public static boolean useHeuristicUpperbound = true;

    public static int solve(Graph g, int t) {
        long startTime = System.currentTimeMillis();
        initSolver(g, t);
        log("heu res: " + heuristicRes);
        log("init lb: " + initLB);
        while (true) {
            if (budget == resWith) break;
            log("#try k " + budget);
            if (branch(0)) {//found solution
                break;
            } else {//no solution found -> increase cost budget
                noResWith = budget;
                budget++;
            }
        }
        log("t: " + t);
        log("k: " + resWith);
        log("rec steps: " + branchTreeNodeCounter);
        log("lb1Count: " + lb1Count);
        log("lb2Count: " + lb2Count);

//        Clustering.printInformation(); // to see the cluster
        System.out.println("time-java: " + (System.currentTimeMillis() - startTime));
        return resWith;
    }

    private static void initSolver(Graph g, int t) {
        branchTreeNodeCounter = 0;
        lb1Count = 0;
        lb2Count = 0;

        Branch.g = g;
        Branch.t = t;

        if (useHeuristicUpperbound) {
            resWith = Heuristic.computeUpperBound(g, t);
            heuristicRes = resWith;
        } else resWith = Integer.MAX_VALUE;

        g.initializeNonCommons();
        g.initializeNCsDataStructure();

        if (t >= g.ncCount()) {
            resWith = 0;
            budget = 0;
            noResWith = -1;
            return;
        }

        Packing.updateEverything = true;
        int k = lowerBound(Integer.MAX_VALUE);
        initLB = k;
        noResWith = k - 1;

        if (useHeuristicUpperbound) budget = resWith - 1;//start from top
        else budget = noResWith + 1;//start from bottom
    }

    private static boolean branch(int cost) {
        branchTreeNodeCounter++;
        if (cost > budget) return false;
        if (g.ncCount() <= t) {
            resWith = cost;
            budget = cost - 1;
            log("#new upper bound: " + resWith);
            log("#try " + budget);
            Clustering.savClustering(g);
            return budget == noResWith;//stop search if budget equals noResWith -> current res is optimal
        }

        int k = budget - cost;
        if (lowerBound(k) > k) return false;
        List<Vertex> branchedVertices = g.getBranchVertices(t);
        if (k <= 0) return false;

        List<Edge> addedApartEdges = new LinkedList<>();

        for (int i = 0; i < branchedVertices.size(); i++) {
            Vertex u = branchedVertices.get(i);
            for (int j = i + 1; j < branchedVertices.size(); j++) {
                Vertex v = branchedVertices.get(j);
                Edge e = g.getEdge(u, v);
                if (e.isApart || e.getNumberOfNonCommons() > k || notMergeAble(e)) continue;
                if (branchResolveAllNonCommons(e, cost)) return true;
                e.setApartTrue();
                addedApartEdges.add(e);
                e.u.apartEdgesCount++;
                e.v.apartEdgesCount++;
            }
        }

        for (Edge apartEdge : addedApartEdges) {
            apartEdge.setApartFalse();
            apartEdge.u.apartEdgesCount--;
            apartEdge.v.apartEdgesCount--;
        }
        return false;
    }

    private static boolean branchResolveAllNonCommons(Edge e, int cost) {
        branchTreeNodeCounter++;
        int k = budget - cost;
        if (e.getNumberOfNonCommons() == 0) {
            //reduction rule: resolveNonCommonsOfMergedEdges
            Stack<Edge> editedEdges = resolveNonCommonsOfMergedEdges(g);
            if (editedEdges == null) return false;
            LinkedList<Edge> blockedEdges = new LinkedList<>();
            for (Edge editedEdge : editedEdges) {
                if (!editedEdge.isBlocked) {
                    editedEdge.block();
                    blockedEdges.add(editedEdge);
                }
            }

            setMerged(e, true);
            //branch
            if (branch(cost + editedEdges.size())) return true;
            setMerged(e, false);

            for (Edge blockedEdge : blockedEdges) blockedEdge.unblock();
            for (Edge editedEdge : editedEdges) g.editEdge(editedEdge);
            return false;
        }
        //resolve a non common neighbor/witness
        if (k <= 0) return false;
        if (lowerBound(k) > k) return false;

        Vertex nonCommon = e.nonCommons.getLastElement();

        Edge e1 = g.getEdge(e.u, nonCommon);
        boolean branchA = !e1.isBlocked;
        if (branchA) {
            g.editEdge(e1);
            e1.block();
            if (branchResolveAllNonCommons(e, cost + 1)) return true;
            g.editEdge(e1);
        }

        Edge e2 = g.getEdge(e.v, nonCommon);
        boolean branchB = !e2.isBlocked;
        if (branchB) {
            g.editEdge(e2);
            e2.block();
            if (branchResolveAllNonCommons(e, cost + 1)) return true;
            g.editEdge(e2);
        }

        if (branchB) e2.unblock();
        if (branchA) e1.unblock();
        return false;
    }

    private static boolean notMergeAble(Edge e) {
        Vertex v = e.v;
        Vertex u = e.u;
        for (Vertex w : e.nonCommons) {
            if (g.getEdge(u, w).isBlocked && g.getEdge(v, w).isBlocked) return true;
        }
        return false;
    }

    public static Stack<Edge> resolveNonCommonsOfMergedEdges(Graph g) {
        Stack<Edge> editedEdges = new Stack<>();
        boolean progress = true;
        while (progress) {
            progress = false;
            for (Edge edge : mergedEdges) {
                if (edge.getNumberOfNonCommons() != 0) {
                    Vertex u = edge.u;
                    Vertex v = edge.v;
                    while (edge.getNumberOfNonCommons() != 0) {
                        Vertex nonCommon = edge.nonCommons.getLastElement();
                        Edge un = g.getEdge(u, nonCommon);
                        Edge vn = g.getEdge(v, nonCommon);
                        boolean unNotEditAble = un.isEdited || un.isBlocked;
                        boolean vnNotEditAble = vn.isEdited || vn.isBlocked;

                        if (unNotEditAble && vnNotEditAble) {
                            //undo changes and return null as signal to cut the branch off
                            for (Edge editedEdge : editedEdges) g.editEdge(editedEdge);
                            return null;
                        }
                        if (unNotEditAble) {
                            g.editEdge(vn);
                            editedEdges.add(vn);
                            progress = true;
                        }
                        if (vnNotEditAble) {
                            g.editEdge(un);
                            editedEdges.add(un);
                            progress = true;
                        }
                    }
                }
            }
        }
        return editedEdges;
    }

    private static int lowerBound(int k) {
        int lb = LowerBound.lb1(g, t);
        if (lb > k) {
            lb1Count++;
            return lb;
        }
        lb = LowerBound.lb2(g, t, k, false);
        if (lb > k) lb2Count++;
        return lb;
    }

    private static void setMerged(Edge e, boolean newValue) {
        if (newValue) mergedEdges.push(e);
        else mergedEdges.pop();
        e.isMerged = newValue;
    }
}