package algo.branch;

import algo.datastructure.*;
import algo.heuristic.Heuristic;

import java.util.*;

import static algo.branch.LowerBound.lb1Count;
import static algo.branch.LowerBound.lb2Count;
import static algo.util.Util.log;

public class BranchMerge {
    static int branchTreeNodeCounter;

    static Graph g;
    private static int t;

    static int resWith;
    static int noResWith;
    static int budget;

    public static int initLB = 0;
    private static int heuristicRes = -1;

    static Comparator<Vertex> comparatorVertexWeight = Comparator.comparingInt(vertex -> vertex.weight);

    //settings
    public static boolean useReduction = true;
    public static boolean useModifiedBranching = true;
    public static boolean useHeuristicUpperbound = false;

    public static int solve(Graph g, int t) {
        initSolver(g, t);
        log("heu res: " + heuristicRes);
        log("init lb: " + initLB);

        while (true) {
            if (budget == resWith) break;
            log("#try k " + budget);
            if (branch(0)) {//found solution
                break;
            } else {//no solution found -> increase cost budget
                noResWith = budget;
                budget++;
            }
        }

        log("t: " + t);
        log("k: " + resWith);
        log("rec steps: " + branchTreeNodeCounter);
        log("lb1Count: " + lb1Count);
        log("lb2Count: " + lb2Count);
        return resWith;
    }

    private static void initSolver(Graph g, int t) {
        branchTreeNodeCounter = 0;
        lb1Count = 0;
        lb2Count = 0;

        BranchMerge.g = g;
        BranchMerge.t = t;

        if (useHeuristicUpperbound) {
            resWith = Heuristic.computeUpperBound(g, t);
            heuristicRes = resWith;
        } else resWith = Integer.MAX_VALUE;

        if (useReduction) NC.initializeNCs(g);
        g.initializeNonCommons();
        g.initializeNCsDataStructure();

        if (t >= g.ncCount()) {
            resWith = 0;
            budget = 0;
            noResWith = -1;
            return;
        }

        Packing.updateEverything = true;
        int k = lowerBound(Integer.MAX_VALUE);
        initLB = k;
        noResWith = k - 1;

        if (useHeuristicUpperbound) budget = resWith - 1;//start from top
        else budget = noResWith + 1;//start from bottom
    }

    static boolean branch(int cost) {
        branchTreeNodeCounter++;
        int k = budget - cost;

        if (lowerBound(k) > k) return false;
        if (k < 0) return false;
        g.selectFirstVertices(t + 1, comparatorVertexWeight);
        List<Vertex> branchVertices = g.getBranchVertices(t);

        if (branchVertices.size() <= t) {//solution found with budget
            resWith = cost;
            budget = cost - 1;
            log("#new upper bound: " + resWith);
            log("#try " + budget);
            return budget == noResWith;//stop search if budget equals noResWith -> current res is optimal
        }

        Stack<Edge> addedApartEdges = new Stack<>();

        for (int i = 0; i < branchVertices.size(); i++) {
            Vertex u = branchVertices.get(i);
            for (int j = i + 1; j < branchVertices.size(); j++) {
                Vertex v = branchVertices.get(j);
                Edge edge = g.getEdge(u, v);

                if (edge.isApart || notMergeAble(edge, k)) continue;

                if (useModifiedBranching && g.haveSameAdjacencyInsideAndTowards(u, v)) {
                    if (branchResolveAllNonCommons(edge, u, v, cost)) return true;
                } else {
                    if (branchMergeClique(cost, u, v, edge, true)) return true;
                    if (branchMergeClique(cost, u, v, edge, false)) return true;
                }

                edge.setApartTrue();
                u.apartEdgesCount++;
                v.apartEdgesCount++;
                addedApartEdges.add(edge);
            }
        }
        for (Edge apartEdge : addedApartEdges) {
            apartEdge.setApartFalse();
            apartEdge.u.apartEdgesCount--;
            apartEdge.v.apartEdgesCount--;
        }
        return false;
    }

    private static boolean branchMergeClique(int cost, Vertex u, Vertex v, Edge edge, boolean shouldBeClique) {
        if (edge.isEdited && (edge.isEdge != shouldBeClique)) return false;
        boolean falseLabel = (shouldBeClique) ?
                u.label == Label.SEPARATE || v.label == Label.SEPARATE :
                u.label == Label.ADJACENT || v.label == Label.ADJACENT;
        if (falseLabel) return false;

        int oldCost = u.computeInsideCost() + v.computeInsideCost();//order: first compute,then set new label
        boolean undoLabelU = setLabel(u, shouldBeClique);
        boolean undoLabelV = setLabel(v, shouldBeClique);

        int costIncreaseByMergeUV = u.computeInsideCost() + v.computeInsideCost() - oldCost;

        //begin: reduction rule
        boolean reductionApplied = false;
        if (useReduction) {
            if (SameNCReductionRule.apply(u, v, shouldBeClique, cost + costIncreaseByMergeUV))
                return true;
            reductionApplied = SameNCReductionRule.appliedFlag;//if true, then reduction rule was applied and returned false
        }
        //end: reduction rule

        //branch if reduction rule did not applied
        if (!reductionApplied) {
            boolean editEdge = edge.isEdge != shouldBeClique;
            if (editEdge) {
                costIncreaseByMergeUV += edge.getEditCost();
                g.editEdge(edge);
            }

            if (branchResolveAllNonCommons(edge, u, v, cost + costIncreaseByMergeUV)) return true;
            if (editEdge) g.editEdge(edge);
        }

        if (undoLabelV) undoLabel(v);
        if (undoLabelU) undoLabel(u);

        return false;
    }

    private static boolean branchResolveAllNonCommons(Edge edge, Vertex u, Vertex v, int cost) {
        branchTreeNodeCounter++;

        int k = budget - cost;
        if (lowerBound(k) > k) return false;
        if (k < 0) return false;
        if (edge.getNumberOfNonCommons() == 0) {
            //set label correctly
            Label labelU = u.label;
            Label labelV = v.label;
            if (labelU == Label.UNDECIDED) u.label = v.label;
            if (labelV == Label.UNDECIDED) v.label = u.label;
            //merge
            Edge[] revertNonCommons = merge(u, v);

            if (branch(cost)) return true;

            u.label = labelU;
            v.label = labelV;

            apart(u, v, revertNonCommons);
            return false;
        }

        Vertex nonCommon = edge.nonCommons.getLastElement();//get last nonCommon of array
        boolean branchA;
        boolean branchB;
        Edge e1 = g.getEdge(u, nonCommon);
        if (branchA = !e1.isBlocked) {
            g.editEdge(e1);
            e1.block();
            if (branchResolveAllNonCommons(edge, u, v, cost + e1.getEditCost())) return true;
            g.editEdge(e1);
        }

        Edge e2 = g.getEdge(v, nonCommon);
        if (branchB = !e2.isBlocked) {
            g.editEdge(e2);
            e2.block();
            if (branchResolveAllNonCommons(edge, u, v, cost + e2.getEditCost())) return true;
            g.editEdge(e2);
        }

        if (branchB) e2.unblock();
        if (branchA) e1.unblock();

        return false;
    }

    static boolean setLabel(Vertex v, boolean label) {
        boolean set = v.label == Label.UNDECIDED;
        if (set) {
            Label newLabel = label ? Label.ADJACENT : Label.SEPARATE;
            Packing.updatePackingByChangeLabel(v, v.label, newLabel);

            v.label = newLabel;
            g.searchNCAndMoveVertex(v);
        }

        return set;
    }

    static void undoLabel(Vertex v) {//only call method if it was set
        //order is here important
        Packing.updatePackingByChangeLabel(v, v.label, Label.UNDECIDED);
        v.label = Label.UNDECIDED;
        g.searchNCAndMoveVertex(v);
    }

    static int lowerBound(int k) {
        if ((g.ncCount() - t + 1) / 2 > k) return k + 1;
        if (g.ncCount() <= t) return 0;
        int lb = LowerBound.lb1(g, t);
        if (lb > k) {
            lb1Count++;
            return lb;
        }
        lb = LowerBound.lb2(g, t, k, false);
        if (lb > k) lb2Count++;
        return lb;
    }

    public static Edge[] merge(Vertex u, Vertex v) {
        Edge mergeEdge = g.getEdge(u, v);
        g.removeVertex(v);

        u.insideEdges += v.insideEdges;
        if (mergeEdge.isEdge) u.insideEdges += (u.weight * v.weight);
        u.weight += v.weight;

        for (Vertex neighbor : v) g.removeEdgeOneDirectional(neighbor, v);

        Edge[] nonCommonWithN2 = new Edge[v.nonCommonForTheseEdges.size()];
        v.nonCommonForTheseEdges.toArray(nonCommonWithN2);

        for (Edge edge : nonCommonWithN2) edge.removeNonCommon(v);
        for (Edge edge : u.nonCommonForTheseEdges)
            edge.vertexWeightsOfNonCommons += v.weight;

        mergeEdge.isMerged = true;

        NC originNC = v.currentNC;
        originNC.removeVertex(v);
        g.ncs.updateOfEmptyNCByRemove(originNC);

        u.currentNC.size += v.weight;

        g.searchNCAndMoveVertex(u);

        v.setParentVertex(u);
        Packing.updatePackingByMerge();
        return nonCommonWithN2;
    }

    public static void apart(Vertex u, Vertex v, Edge[] revertNonCommons) {
        Edge mergedEdge = g.getEdge(u, v);
        mergedEdge.isMerged = false;

        for (Edge edge : revertNonCommons) edge.addNonCommon(v);
        for (Edge edge : u.nonCommonForTheseEdges)
            edge.vertexWeightsOfNonCommons -= v.weight;

        for (Vertex neighbor : v) g.addEdgeOneDirectional(neighbor, v);

        u.weight -= v.weight;
        if (g.getEdge(u, v).isEdge) u.insideEdges -= (u.weight * v.weight);
        u.insideEdges -= v.insideEdges;

        u.currentNC.size -= v.weight;
        g.searchNCAndMoveVertex(u);
        g.searchNCAndMoveVertex(v);

        g.addVertex(v);
        v.removeLastParentVertex();
        Packing.updatePackingByApart();
    }

    public static boolean notMergeAble(Edge edge, int k) {
        if (edge.isNotMergeableWithLabel()) return true;

        int minCost = 0;
        for (Vertex nonCommon : edge.nonCommons) {
            Edge e1 = g.getEdge(edge.u, nonCommon);
            Edge e2 = g.getEdge(edge.v, nonCommon);
            if (e1.isBlocked && e2.isBlocked) return true;
            if (e1.isBlocked) {
                minCost += e2.getEditCost();
            } else {
                if (e2.isBlocked) minCost += e1.getEditCost();
                else minCost += Math.min(e1.getEditCost(), e2.getEditCost());
            }
            if (minCost > k) return true;
        }
        return false;
    }
}
