package algo.branch;

import algo.datastructure.Edge;
import algo.datastructure.Graph;
import algo.datastructure.PartSetVertices;
import algo.datastructure.Vertex;

public class Pack {

    private final int id;
    public Vertex[] packedVertices;
    public int[] indicesOfPacksArrayOfVertex;
    int cost;
    int size = 0;

    static private int idCounter = 0;

    boolean updatePackCount = false;
    boolean isDeleted = false;

    Pack(int t) {
        packedVertices = new Vertex[t + 1];
        indicesOfPacksArrayOfVertex = new int[t + 1];
        id = idCounter++;
    }

    public void addVertex(Vertex v) {
        packedVertices[size] = v;
        indicesOfPacksArrayOfVertex[size] = v.inManyPacks;
        v.expandBlockArrayIfNecessary();
        v.packs[v.inManyPacks] = this;
        v.indicesInPacks[v.inManyPacks] = size;
        size++;
        v.inManyPacks++;
    }

    void removeVertex(Vertex packedVertex) {
        swapPositionsWithLastEntry(packedVertex);
        packedVertex.swapPackWithLastPack(indicesOfPacksArrayOfVertex[size - 1]);
        packedVertex.removeLastPack();
        size--;
        packedVertices[size] = null;
        indicesOfPacksArrayOfVertex[size] = -1;
    }

    void computeCost(Graph g) {
        int min = Integer.MAX_VALUE;
        for (int i = 0; i < size; i++) {
            Vertex u = packedVertices[i];
            for (int j = i + 1; j < size; j++) {
                Vertex v = packedVertices[j];
                Edge e = g.getEdge(u, v);
                if (e.isApart || e.isNotMergeableWithLabel()) continue;
                min = Math.min(e.getResolveCostOfNonCommons(), min);
            }
        }
        cost = min;
    }

    Edge findBestStartEdge(Graph g, PartSetVertices unpackedVertices) {
        int maxValue = 0;
        Edge bestEdge = null;
        for (int i = 0; i < unpackedVertices.size(); i++) {
            Vertex u = unpackedVertices.get(i);
            for (int j = i + 1; j < unpackedVertices.size(); j++) {
                Vertex v = unpackedVertices.get(j);
                Edge e = g.getEdge(v, u);
                if (g.areSimilar(u, v) || e.isApart || e.isNotMergeableWithLabel()) continue;
                //weighted variant delivered better results
                int value = e.vertexWeightsOfNonCommons
                        * Math.min(u.weight, v.weight);
                //unweighted variant:
//                int value = e.getCostOfWitnesses();
                if (value > maxValue) {
                    maxValue = e.getResolveCostOfNonCommons();
                    bestEdge = e;
                }
            }
        }
        if (bestEdge != null) {
            addVertex(bestEdge.u);
            addVertex(bestEdge.v);
            cost = maxValue;
        }
        return bestEdge;
    }

    int minConnectionToPack(Graph g, Vertex vertex) {
        int minValue = Integer.MAX_VALUE;
        for (int i = 0; i < size; i++) {
            Vertex packVertex = packedVertices[i];
            if (g.areSimilar(vertex, packVertex)) return -1;//are in one neighborhood class

            Edge e = g.getEdge(vertex, packVertex);
            if (e.isApart || e.isNotMergeableWithLabel()) continue;
            if (e.getResolveCostOfNonCommons() < minValue)
                minValue = e.getNumberOfNonCommons();
        }
        return minValue;
    }

    void swapPositionsWithLastEntry(Vertex v) {
        int posVSwap = 0;
        //find position of v
        for (int i = 0; i < packedVertices.length; i++) {
            Vertex packedVertex = packedVertices[i];
            if (v == packedVertex) {
                posVSwap = i;
                break;
            }
        }

        //swap
        int lastEntry = size-1;
        Vertex auxVertex = packedVertices[posVSwap];
        int auxIndex = indicesOfPacksArrayOfVertex[posVSwap];

        packedVertices[posVSwap] = packedVertices[lastEntry];
        indicesOfPacksArrayOfVertex[posVSwap] = indicesOfPacksArrayOfVertex[lastEntry];

        packedVertices[lastEntry] = auxVertex;
        indicesOfPacksArrayOfVertex[lastEntry] = auxIndex;

        packedVertices[lastEntry].indicesInPacks[indicesOfPacksArrayOfVertex[lastEntry]] = lastEntry;
        packedVertices[posVSwap].indicesInPacks[indicesOfPacksArrayOfVertex[posVSwap]] = posVSwap;
    }

    @Override
    public boolean equals(Object o) {
        return this == o;
    }

    @Override
    public int hashCode() {
        return id;
    }

}
