package algo.datastructure;

import algo.util.Util;

import java.util.*;

import static algo.util.Util.over2;

public class BlockPartition {

    private final List<Block> blocks;
    private int[][] edgesBetweenBlocks;
    private int[][] edgesFromVertexToBlock;

    public BlockPartition(List<NC> ncs) {
        blocks = new ArrayList<>(ncs.size());
        for (int i = 0; i < ncs.size(); i++) {
            Block b = new Block(ncs.size(), i);
            blocks.add(b);
            b.addAllVerticesOfNC(ncs.get(i));
        }
        setBlockIndices();
        edgesBetweenBlocks = computeEdgesBetweenBlocksAndSet();
    }

    public void initializeUseOfEdgesFromVertexToBlock(Graph graph) {
        edgesFromVertexToBlock = new int[graph.vertexCount][blocks.size()];
        for (int i = 0; i < graph.vertexCount; i++) {
            Vertex vertex = graph.get(i);
            for (Vertex neighbor : vertex) {
                edgesFromVertexToBlock[vertex.id][neighbor.getBlockIndex()]++;
            }
        }
    }

    private int[][] computeEdgesBetweenBlocksAndSet() {
        int maxIndex = blocks.stream().mapToInt(b -> b.index).max().getAsInt() + 1;
        int[][] edgesBetweenBlock = new int[maxIndex][maxIndex];
        for (Block block : blocks) {
            for (int i = 0; i < block.size; i++) {
                Vertex vertex = block.verticesArray[i];
                for (Vertex neighbor : vertex)
                    edgesBetweenBlock[block.index][neighbor.currentBlock.index]++;
            }
            edgesBetweenBlock[block.index][block.index] /= 2;
        }
        this.edgesBetweenBlocks = edgesBetweenBlock;
        return edgesBetweenBlock;
    }

    public void merge(int blockIndex1, int blockIndex2) {
        Block blockToFill = blocks.get(blockIndex1);
        List<Vertex> moveVertices = new ArrayList<>(Arrays.asList(blocks.get(blockIndex2).verticesArray).subList(0, blocks.get(blockIndex2).size));
        for (Vertex vertex : moveVertices)
            moveVertexToBlock(vertex, blockToFill);
        blocks.remove(blockIndex2);
    }

    public void moveVertexToBlock(Vertex vertex, Block destinationBlock) {
        Block currentBlock = vertex.currentBlock;
        for (Vertex neighbor : vertex) {
            int neighborBlockIndex = neighbor.getBlockIndex();

            if (edgesFromVertexToBlock != null) {
                edgesFromVertexToBlock[neighbor.id][destinationBlock.index]++;
                edgesFromVertexToBlock[neighbor.id][currentBlock.index]--;
            }

            edgesBetweenBlocks[neighborBlockIndex][destinationBlock.index]++;
            if (neighborBlockIndex != destinationBlock.index)
                edgesBetweenBlocks[destinationBlock.index][neighborBlockIndex]++;

            edgesBetweenBlocks[neighborBlockIndex][currentBlock.index]--;
            if (neighborBlockIndex != currentBlock.index)
                edgesBetweenBlocks[currentBlock.index][neighborBlockIndex]--;
        }
        //remove first, then add
        currentBlock.removeVertex(vertex);
        destinationBlock.addVertex(vertex);
    }

    public void splitBlock(Block block) {
        List<Vertex> moveVertices = new ArrayList<>(Arrays.asList(block.verticesArray).subList(0, block.size));
        for (Vertex vertex : moveVertices) {
            BestBlocksResult res = computeBestBlocksByMovingVertex(vertex);
            Block newBlock = res.getRandomBestBlock();
            moveVertexToBlock(vertex, newBlock);
        }
        blocks.remove(block);
    }

    public BestBlocksResult computeBestBlocksByMovingVertex(Vertex vertex) {
        int minCostIncrease = Integer.MAX_VALUE;
        Set<Block> bestBlocks = new HashSet<>();
        Block currentBlock = vertex.currentBlock;

        for (Block block : blocks) {
            if (currentBlock == block) continue;
            int costIncreaseOfMovingVertexInBlock = computeCostIncreaseByMoveVertexToBlock(vertex, block);

            if (costIncreaseOfMovingVertexInBlock < minCostIncrease) {
                minCostIncrease = costIncreaseOfMovingVertexInBlock;
                bestBlocks.clear();
            }
            if (minCostIncrease == costIncreaseOfMovingVertexInBlock)
                bestBlocks.add(block);
        }

        return new BestBlocksResult(bestBlocks, minCostIncrease);
    }

    private int computeCostIncreaseByMoveVertexToBlock(Vertex vertex, Block block) {
        Block currentBlock = vertex.currentBlock;
        int costIncrease = 0;
        for (Block otherBlock : blocks) {//compute cost-increase my moving to otherBlock
            if (otherBlock == currentBlock) continue;
            int currentEdgesBetweenBlocks = getEdgeCountBetween(block, otherBlock);
            int currentEdgesFromVertexToOtherBlock = edgesFromVertexToBlock[vertex.id][otherBlock.index];

            int currentMaxEdgesCount;
            int futureMaxEdgesCount;
            if (block == otherBlock) {
                currentMaxEdgesCount = over2(block.size);
                futureMaxEdgesCount = over2(block.size + 1);
            } else {
                currentMaxEdgesCount = block.size * otherBlock.size;
                futureMaxEdgesCount = (block.size + 1) * otherBlock.size;
            }

            int currentCost = Math.min(currentMaxEdgesCount - currentEdgesBetweenBlocks, currentEdgesBetweenBlocks);
            int futureCost = Math.min(futureMaxEdgesCount - (currentEdgesBetweenBlocks + currentEdgesFromVertexToOtherBlock)
                    , (currentEdgesBetweenBlocks + currentEdgesFromVertexToOtherBlock));

            costIncrease += (futureCost - currentCost);
        }
        return costIncrease;
    }

    public int computeCostOfBlock(Block block) {
        int cost = 0;
        cost += getCostWithin(block);
        for (Block otherBlock : blocks) {
            if (otherBlock == block) continue;
            cost += getCostBetween(block, otherBlock);
        }
        return cost;
    }

    public int getCostWithin(Block block) {
        int edgesWithin = getEdgeCountBetween(block, block);
        return Math.min(over2(block.size) - edgesWithin, edgesWithin);
    }

    public int getCostBetween(Block blockA, Block blockB) {
        int edgesBetween = getEdgeCountBetween(blockA, blockB);
        return Math.min(blockA.size * blockB.size - edgesBetween, edgesBetween);
    }

    public Block addEmptyBlock(int initialNCSize) {
        Block newBlock = new Block(initialNCSize, blocks.size());
        blocks.add(newBlock);
        setBlockIndices();
        computeEdgesBetweenBlocksAndSet();
        return newBlock;
    }

    public void removeEmptyBlock(Block auxBlock) {
        //this works only because newBlock of method "addEmptyBlock" is the last element! Otherwise, indices need to be adapted
        blocks.remove(auxBlock);
    }

    private void setBlockIndices() {
        for (int i = 0; i < blocks.size(); i++)
            blocks.get(i).index = i;
    }

    public int getEdgeCountBetween(Block blockA, Block blockB) {
        return edgesBetweenBlocks[blockA.index][blockB.index];
    }

    public int getEdgeCountBetween(Vertex vertex, Block block) {
        return edgesFromVertexToBlock[vertex.id][block.index];
    }

    public Block getBlock(int index) {
        return blocks.get(index);
    }

    public List<Block> getBlocks() {
        return blocks;
    }

    public int size() {
        return blocks.size();
    }

    public static class BestBlocksResult {
        public final Set<Block> bestBlocks;
        public final int minCost;

        public BestBlocksResult(Set<Block> bestBlocks, int minCost) {
            this.bestBlocks = bestBlocks;
            this.minCost = minCost;
        }

        public Block getRandomBestBlock() {
            return Util.getRandomElementOf(bestBlocks);
        }
    }

    public int computeCost() {
        int cost = 0;
        for (int i = 0; i < blocks.size(); i++) {
            Block b1 = blocks.get(i);
            for (int j = i + 1; j < blocks.size(); j++) {
                Block b2 = blocks.get(j);
                int maxEdges = b1.size() * b2.size();
                int edgesBetween = edgesBetweenBlocks[b1.index][b2.index];
                cost += Math.min(maxEdges - edgesBetween, edgesBetween);
            }
        }

        for (Block b : blocks) {
            int edgesWithin = edgesBetweenBlocks[b.index][b.index];
            cost += Math.min(over2(b.size()) - edgesWithin, edgesWithin);
        }
        return cost;
    }
}
