package algo.datastructure;

import algo.util.QuickSelect;

import java.util.*;
import java.util.function.Consumer;

public class PartSetVertices implements Collection<Vertex> {
    private Vertex[] vertices;
    int[] indices;
    int size = 0;

    public PartSetVertices(Graph g) {
        this(g.initialVertexCount);
    }

    PartSetVertices(int maximalNumberOfVertices) {
        indices = new int[maximalNumberOfVertices];
        Arrays.fill(indices, -1);
        vertices = new Vertex[10];
    }

    @Override
    public boolean add(Vertex vertex) {
        if (indices[vertex.id] != -1) return false;
        expandIfNecessary();
        vertices[size] = vertex;
        indices[vertex.id] = size;
        size++;
        return true;
    }

    public boolean remove(Vertex vertex) {
        int index = indices[vertex.id];
        if (index == -1) return false;
        size--;
        if (size != index) {
            Vertex changedVertex = vertices[size];
            vertices[index] = changedVertex;
            indices[changedVertex.id] = index;
        }
        vertices[size] = null;
        indices[vertex.id] = -1;
        return true;
    }

    public Vertex getLastElement(){
        return vertices[size - 1];
    }

    void expandIfNecessary() {
        if (size == vertices.length) {
            Vertex[] old = vertices;
            vertices = new Vertex[(int) (size * 1.5)];
            System.arraycopy(old, 0, vertices, 0, size);
        }
    }

    public void shuffle(){
        Random r = new Random();
        for (int i = size - 1; i > 0; i --){
            int j = r.nextInt( i + 1);
            Vertex aux = vertices[i];
            vertices[i] = vertices[j];
            vertices[j] = aux;
        }
        for (int i = 0; i < size; i++) {
            Vertex vertex = vertices[i];
            indices[vertex.id] = i;
        }
    }

    public void selectFirstVertices(int numberToSelect, Comparator<Vertex> comparator) {
        QuickSelect.selectSortTopKMax(vertices, numberToSelect, size, comparator);
        for (int i = 0; i < size; i++) indices[vertices[i].id] = i;
    }

    public Vertex get(int index ){
        return vertices[index];
    }

    @Override
    public boolean remove(Object o) {
        return remove((Vertex) o);
    }

    @Override
    public int size() {
        return size;
    }

    @Override
    public boolean isEmpty() {
        return size == 0;
    }

    public boolean contains(Vertex vertex) {
        return indices[vertex.id] != -1;
    }

    @Override
    public boolean contains(Object o) {
        return indices[((Vertex) o).id] != -1;
    }

    @Override
    public Iterator<Vertex> iterator() {
        return new Iterator<>() {
            int index = 0;
            @Override
            public boolean hasNext() {
                return index != size;
            }

            @Override
            public Vertex next() {
                return vertices[index++];
            }
        };
    }

    @Override
    public void forEach(Consumer<? super Vertex> action) {
        for (int i = size - 1; i >= 0; i--) {
            Vertex vertex = vertices[i];
            action.accept(vertex);
        }
    }

    @Override
    public Object[] toArray() {
        return Arrays.copyOf(vertices, size);
    }

    @Override
    public <T> T[] toArray(T[] a) {
        return (T[]) Arrays.copyOf(vertices, size, a.getClass());
    }

    @Override
    public boolean containsAll(Collection<?> c) {
        return c.stream().allMatch(this::contains);
    }

    @Override
    public boolean addAll(Collection<? extends Vertex> c) {
        int oldSize = size;
        c.forEach(this::add);
        return oldSize != size;
    }

    @Override
    public boolean removeAll(Collection<?> c) {
        int oldSize = size;
        c.forEach(this::remove);
        return oldSize != size;
    }

    @Override
    public boolean retainAll(Collection<?> c) {
        int oldSize = size;
        for (Object o : c) if (!contains(o)) remove(o);
        return oldSize != size;
    }

    @Override
    public void clear() {
        for (int i = size - 1; i >= 0; i--) remove(vertices[i]);
    }

}
