package algo.heuristic;

import algo.datastructure.*;
import algo.util.Util;

import java.util.LinkedList;
import java.util.List;

public class MergeHeuristic extends Util {

    private static final boolean improveByLocalSearch = true;

    public static BlockPartitionResult solve(Graph graph, int t) {
        long start = System.currentTimeMillis();


        List<NC> initialsNCs = NC.initializeNCs(graph);

        BlockPartition blockPartition = new BlockPartition(initialsNCs);

        while (blockPartition.size() > t) {
            int[] minBags = findBestMergeBags(blockPartition);
            blockPartition.merge(minBags[0], minBags[1]);
        }

        long greedy = (System.currentTimeMillis()) - start;
        long timeStamp = System.currentTimeMillis();
        int res;
        if (improveByLocalSearch) res = LocalSearch.improve(graph, blockPartition, initialsNCs);
        else res = blockPartition.computeCost();

        for (Vertex vertex : graph) {
            vertex.currentBlock = null;
            vertex.indexOfBlockArrayPosition = -1;
        }

        long localSearch = System.currentTimeMillis() - timeStamp;
//        System.out.println("only merge "  + greedy + ", local search " + localSearch);

        return new BlockPartitionResult(blockPartition, res);
    }

    private static int[] findBestMergeBags(BlockPartition blockPartition) {
        int minCost = Integer.MAX_VALUE;
        List<int[]> bestPartners = new LinkedList<>();

        for (int i = 0; i < blockPartition.size(); i++) {
            Block block1 = blockPartition.getBlock(i);
            for (int j = i + 1; j < blockPartition.size(); j++) {//merge i and j
                Block block2 = blockPartition.getBlock(j);

                int costIncrease = getCostIncreaseByMerge(blockPartition, block1, block2);
                if (costIncrease < minCost) {
                    minCost = costIncrease;
                    bestPartners.clear();
                }
                if (costIncrease == minCost) bestPartners.add(new int[]{i, j});
            }
        }
        return Util.getRandomElementOf(bestPartners);
    }

    private static int getCostIncreaseByMerge(BlockPartition blockPartition, Block block1, Block block2) {
        int costIncrease = 0;

        int bagSizeMerged = block1.size() + block2.size();
        //cost for intern edges
        int currentCostWithin = blockPartition.getCostWithin(block1) + blockPartition.getCostWithin(block2);

        int edgesWithinMerged = blockPartition.getEdgeCountBetween(block1, block1)
                + blockPartition.getEdgeCountBetween(block2, block2)
                + blockPartition.getEdgeCountBetween(block1, block2);

        costIncrease += Math.min(over2(bagSizeMerged) - edgesWithinMerged, edgesWithinMerged) - currentCostWithin;

        //cost to other blocks
        for (int b = 0; b < blockPartition.size(); b++) {
            Block bB = blockPartition.getBlock(b);
            if (bB == block1 || bB == block2) continue;
            int bagSizeB = bB.size();
            int edgesBetweenIAndB = blockPartition.getEdgeCountBetween(block1, bB);
            int edgesBetweenJAndB = blockPartition.getEdgeCountBetween(block2, bB);

            int currentCost = blockPartition.getCostBetween(block1, bB) + blockPartition.getCostBetween(block2, bB);

            int edgesToBagMerged = edgesBetweenIAndB + edgesBetweenJAndB;
            costIncrease += Math.min(bagSizeB * bagSizeMerged - edgesToBagMerged, edgesToBagMerged) - currentCost;
        }
        return costIncrease;
    }

    public static int[] solveSiamTest(Graph graph, int t) {
        List<NC> initialsNCs = NC.initializeNCs(graph);
        BlockPartition blockPartition = new BlockPartition(initialsNCs);

        while (blockPartition.size() > t) {
            int[] minBags = findBestMergeBags(blockPartition);
            blockPartition.merge(minBags[0], minBags[1]);
        }

        int[] res = new int[2];
        res[0] = blockPartition.computeCost();
        res[1] = LocalSearch.improve(graph, blockPartition, initialsNCs);

        for (Vertex vertex : graph) {
            vertex.currentBlock = null;
            vertex.indexOfBlockArrayPosition = -1;
        }
        return res;
    }
}
