package algo.ilp;

import algo.datastructure.*;
import algo.heuristic.BlockPartitionResult;
import algo.heuristic.Heuristic;
import algo.test.Clustering;
import algo.util.Util;
import gurobi.*;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import static algo.util.Util.log;
import static algo.util.Util.over2;

public class ILPV1 {

    private static Graph g;
    private static int t;
    private static GRBVar[][] editArr;
    private static GRBVar[][] mergeEdgeArr;

    public static boolean triadicConstraintsLazy = true; //add triadic constraints lazy
    public static boolean useHeuristicUpperBound = true;

    //set these variables to give Gurobi a time-limit in seconds
    public static boolean useTimeLimit = false;
    public static int timeLimit;
    static int mipSolCallbacks = 0;

    public static int solve(Graph graph, int t) {
        long startTime = System.currentTimeMillis();
        graph.initializeNonCommons();
        ILPV1.g = graph;
        ILPV1.t = t;
        log("t: " + t);
        try {
            GRBEnv env = new GRBEnv(true);
            env.set(GRB.IntParam.Threads, 1);
            env.set(GRB.IntParam.LazyConstraints, 1);
            if (!Util.printLogs) env.set(GRB.IntParam.OutputFlag, 0);//comment if you want more information form Gurobi
            env.start();
            GRBModel model = new GRBModel(env);
            if (useTimeLimit) model.set(GRB.DoubleParam.TimeLimit, timeLimit);

            initializeVariablesAndSetObjective(graph, model);
            //set all merge constraints
            setAllMergeConstraints(graph, model);

            if (!triadicConstraintsLazy) setAllTransitivityConstraints(graph, model); //set all transitivity constraints

            //set callback
            model.setCallback(callback);
            model.optimize();

            int k = countEdits(graph);

            //dispose
            if (model.get(GRB.IntAttr.Status) == GRB.TIME_LIMIT) k = -1;//timeout
            // print clustering for testing, dispose after that
//            Clustering.saveClustering(graph, mergeEdgeArr);
//            Clustering.printInformation();
            //
            model.dispose();
            env.dispose();
            log("mipSolCallbacks: " + mipSolCallbacks);
            System.out.println("time-java: " + (System.currentTimeMillis() - startTime));
            return k;
        } catch (GRBException e) {
            e.printStackTrace();
        }
        System.out.println("time-java: " + (System.currentTimeMillis() - startTime));
        return -1;
    }


    private static void initializeVariablesAndSetObjective(Graph graph, GRBModel model) throws GRBException {
        int vertexCount = graph.vertexCount;
        //creating variables
        editArr = new GRBVar[vertexCount][vertexCount];
        mergeEdgeArr = new GRBVar[vertexCount][vertexCount];
        GRBLinExpr exprMin = new GRBLinExpr();
        for (int i = 0; i < vertexCount; i++) {
            Vertex u = g.get(i);
            for (int j = i + 1; j < vertexCount; j++) {
                Vertex v = g.get(j);

                GRBVar editVar = model.addVar(0.0, 1.0, 0, GRB.BINARY, "edit_" + u.id + "_" + v.id);
                editArr[u.id][v.id] = editVar;
                editArr[v.id][u.id] = editVar;

                GRBVar mergeEdgeVar = model.addVar(0.0, 1.0, 0, GRB.BINARY, "merge_" + u.id + "_" + v.id);
                mergeEdgeArr[u.id][v.id] = mergeEdgeVar;
                mergeEdgeArr[v.id][u.id] = mergeEdgeVar;

                exprMin.addTerm(1.0, editVar);
            }
        }
        //set minimize function
        model.setObjective(exprMin, GRB.MINIMIZE);

        int heuValue = -1;
        if (useHeuristicUpperBound) {
            BlockPartitionResult heuRes = Heuristic.computeUpperBoundWithBlockPartition(g, t);
            heuValue = heuRes.getCost();
            model.addConstr(exprMin, GRB.LESS_EQUAL, heuRes.getCost(), "");
            setStartAssignmentAsHeuristic(heuRes);
        }
        log("heu res: " + heuValue);
    }

    private static void setStartAssignmentAsHeuristic(BlockPartitionResult heuRes) throws GRBException {
        BlockPartition blockPartition = heuRes.getBlockPartition();
        List<Block> blockList = blockPartition.getBlocks();
        for (int i = 0; i < blockList.size(); i++) {
            Block block1 = blockList.get(i);
            for (int j = i; j < blockList.size(); j++) {
                Block block2 = blockList.get(j);

                int edgesBetweenBlocks = blockPartition.getEdgeCountBetween(block1, block2);
                int maxEdges = (i == j) ? over2(block1.size) : block1.size * block2.size;
                boolean connected = edgesBetweenBlocks >= (maxEdges + 1) / 2;

                for (int uIndex = 0; uIndex < block1.size; uIndex++) {
                    Vertex u = block1.get(uIndex);
                    for (int vIndex = (i == j) ? uIndex + 1 : 0; vIndex < block2.size; vIndex++) {
                        Vertex v = block2.get(vIndex);
                        //set editVar
                        editArr[u.id][v.id].set(GRB.DoubleAttr.Start, (g.getEdge(u, v).isEdge != connected) ? 1.0 : 0.0);
                        //set mergeVar
                        mergeEdgeArr[u.id][v.id].set(GRB.DoubleAttr.Start, (i == j) ? 1.0 : 0.0);
                    }
                }
            }
        }
    }

    private static void setAllMergeConstraints(Graph graph, GRBModel model) throws GRBException {
        for (int i = 0; i < graph.vertexCount; i++) {
            Vertex u = graph.get(i);
            for (int j = i + 1; j < graph.vertexCount; j++) {
                Vertex v = graph.get(j);
                List<GRBLinExpr> exprs = getConstraintForMergeEdge(graph, u, v);
                for (GRBLinExpr expr : exprs) {
                    model.addConstr(expr, GRB.LESS_EQUAL, 0, "");
                }
            }
        }
    }

    private static List<GRBLinExpr> getConstraintForMergeEdge(Graph graph, Vertex u, Vertex v) {
        List<GRBLinExpr> res = new LinkedList<>();
        Edge edge = graph.getEdge(u, v);
        GRBVar mergeEdge = mergeEdgeArr[u.id][v.id];

        for (int k = 0; k < graph.vertexCount; k++) {
            Vertex w = graph.get(k);
            if (w == u || w == v) continue;

            GRBLinExpr expr1 = new GRBLinExpr();
            GRBLinExpr expr2 = new GRBLinExpr();
            expr1.addTerm(1.0, mergeEdge);
            expr2.addTerm(1.0, mergeEdge);

            if (edge.containsNonCommonNeighbor(w)) {
                //min one edit:         merge_i_j <= edit_i_n + edit_j_n
                expr1.addTerm(-1.0, editArr[u.id][w.id]);
                expr1.addTerm(-1.0, editArr[v.id][w.id]);
                //max one edit:         merge_i_j <= 2 - edit_i_n - edit_j_n
                expr2.addConstant(-2.0);
                expr2.addTerm(1.0, editArr[u.id][w.id]);
                expr2.addTerm(1.0, editArr[v.id][w.id]);
                res.add(expr1);
                res.add(expr2);
            }
        }
        return res;
    }

    private static void setAllTransitivityConstraints(Graph graph, GRBModel model) throws GRBException {
        for (int i = 0; i < graph.vertexCount; i++) {
            Vertex u = graph.get(i);
            int uIndex = u.id;
            for (int j = i + 1; j < graph.vertexCount; j++) {
                Vertex v = graph.get(j);
                int vIndex = v.id;
                GRBVar mergeEdge = mergeEdgeArr[uIndex][vIndex];
                for (int k = 0; k < graph.vertexCount; k++) {
                    Vertex w = graph.get(k);
                    if (w == u || w == v) continue;
                    GRBVar m1 = mergeEdgeArr[uIndex][w.id];
                    GRBVar m2 = mergeEdgeArr[vIndex][w.id];

                    GRBLinExpr expr = new GRBLinExpr();
                    expr.addConstant(-1.0);
                    expr.addTerm(1.0, m1);
                    expr.addTerm(1.0, m2);
                    model.addConstr(expr, GRB.LESS_EQUAL, mergeEdge, "");
                }
            }
        }
    }

    static GRBCallback callback = new GRBCallback() {
        @Override
        protected void callback() {
            if (where == GRB.Callback.MIPSOL) {
                mipSolCallbacks++;
                try {
                    boolean transitiveConstraintAdded = false;
                    //search for violated transitivity constraints
                    for (int i = 0; i < g.vertexCount; i++) {
                        Vertex v = g.get(i);
                        for (int j = 0; j < g.vertexCount; j++) {
                            if (i == j) continue;
                            Vertex u = g.get(j);
                            if (isMerged(u, v))
                                for (int k = j + 1; k < g.vertexCount; k++) {
                                    if (k == i || k == j) continue;
                                    Vertex w = g.get(k);
                                    if (isMerged(v, w) && !isMerged(u, w)) {
                                        GRBLinExpr expr = new GRBLinExpr();//uv + vw - uw <= 1
                                        expr.addTerm(1.0, mergeEdgeArr[u.id][v.id]);
                                        expr.addTerm(1.0, mergeEdgeArr[w.id][v.id]);
                                        expr.addTerm(-1.0, mergeEdgeArr[u.id][w.id]);
                                        addLazy(expr, GRB.LESS_EQUAL, 1.0);
                                        transitiveConstraintAdded = true;
                                    }
                                }
                        }
                    }
                    if (!transitiveConstraintAdded) {
                        List<Vertex> branchVertices = getBranchNodes();
                        if (branchVertices != null) addLazyBranchConstraint(branchVertices);
                    }
                } catch (GRBException e) {
                    e.printStackTrace();
                }
            }
        }


        private void addLazyBranchConstraint(List<Vertex> branchVertices) throws GRBException {
            addLazy(getLinExprOf(branchVertices), GRB.GREATER_EQUAL, 1);
        }

        private List<Vertex> getBranchNodes() throws GRBException {
            List<Vertex> branchedVertices = new ArrayList<>(t + 1);
            branchedVertices.add(g.get(0));
            for (int i = 1; i < g.vertexCount; i++) {
                Vertex vertex = g.get(i);
                boolean candidate = true;
                for (Vertex branchedVertex : branchedVertices) {
                    if (isMerged(vertex, branchedVertex)) {
                        candidate = false;
                        break;
                    }
                }
                if (candidate) branchedVertices.add(vertex);
                if (branchedVertices.size() > t) return branchedVertices;
            }
            if (branchedVertices.size() > t) return branchedVertices;
            return null;
        }

        boolean isMerged(Vertex u, Vertex v) throws GRBException {
            return equalsOne(getSolution(mergeEdgeArr[u.id][v.id]));
        }
    };


    static GRBLinExpr getLinExprOf(List<Vertex> branchedVertices) {
        GRBLinExpr expr = new GRBLinExpr();
        for (int i = 0; i < branchedVertices.size(); i++) {
            Vertex u = branchedVertices.get(i);
            for (int j = i + 1; j < branchedVertices.size(); j++) {
                Vertex v = branchedVertices.get(j);
                GRBVar mergeEdge = mergeEdgeArr[u.id][v.id];
                expr.addTerm(1.0, mergeEdge);
            }
        }
        return expr;
    }

    static boolean equalsOne(double d) {
        return (int) Math.round(d) == 1;
    }

    static int countEdits(Graph graph) throws GRBException {
        double sum = 0;
        for (int i = 0; i < graph.vertexCount; i++) {
            for (int j = i + 1; j < graph.vertexCount; j++) {
                GRBVar v = editArr[i][j];
                sum += v.get(GRB.DoubleAttr.X);
            }
        }
        int res = (int) Math.round(sum);
        log("k: " + res);
        return res;
    }
}
