package algo.ilp;

import algo.datastructure.*;
import algo.heuristic.BlockPartitionResult;
import algo.heuristic.Heuristic;
import gurobi.*;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import static algo.util.Util.log;
import static algo.util.Util.over2;

public class ILPV2 {

    static Graph g;
    static int t;
    static GRBVar[][] editArr;
    static int[][] initEdgeArr;

    static final boolean useMultipleConstraints = true;
    static final boolean useLimitForMultipleConstraints = true;
    static final int limit = 9;

    public static boolean useHeuristicUpperBound = true;

    //set these variables to give Gurobi a time-limit in seconds
    public static boolean useTimeLimit = false;
    public static int timeLimit;
    static int mipSolCallbacks = 0;

    public static int solve(Graph graph, int t) {
        graph.initializeNonCommons();
        ILPV2.g = graph;
        ILPV2.t = t;

        try {
            GRBEnv env = new GRBEnv(true);
            env.set(GRB.IntParam.Threads, 1);
            env.set(GRB.IntParam.LazyConstraints, 1);
            env.start();

            GRBModel model = new GRBModel(env);
            if (useTimeLimit) model.set(GRB.DoubleParam.TimeLimit, timeLimit);

            InitializeVariablesAndSetObjective(graph, model);

            //add first constraint(s)
            if (useMultipleConstraints) {
                List<GRBLinExpr> exprs = createMultipleConstraints();
                if (exprs != null)
                    for (GRBLinExpr expr : exprs) model.addConstr(expr, GRB.GREATER_EQUAL, 1.0, "");
            } else {
                GRBLinExpr expr = createConstraint();
                if (expr != null) model.addConstr(expr, GRB.GREATER_EQUAL, 1.0, "");
            }

            model.setCallback(callback);
            model.optimize();

            int k = countEdits(graph);

            //dispose
            if (model.get(GRB.IntAttr.Status) == GRB.TIME_LIMIT) k = -1;//timeout
            model.dispose();
            env.dispose();
            log("mipSolCallbacks: " + mipSolCallbacks);

            return k;
        } catch (GRBException e) {
            e.printStackTrace();
        }
        return -1;
    }


    private static void InitializeVariablesAndSetObjective(Graph graph, GRBModel model) throws GRBException {
        int nodeSize = graph.vertexCount;
        //edit_ij
        editArr = new GRBVar[nodeSize][nodeSize];
        initEdgeArr = new int[nodeSize][nodeSize];
        GRBLinExpr exprMin = new GRBLinExpr();
        for (int i = 0; i < nodeSize; i++) {
            Vertex u = graph.get(i);
            for (int j = i + 1; j < nodeSize; j++) {
                Vertex v = graph.get(j);
                if (graph.areNeighbors(u, v)) {
                    initEdgeArr[u.id][v.id] = 1;
                    initEdgeArr[v.id][u.id] = 1;
                }
                GRBVar editVar = model.addVar(0.0, 1.0, 0, GRB.BINARY, "edit_" + i + "_" + j);
                editArr[u.id][v.id] = editVar;
                editArr[v.id][u.id] = editVar;

                exprMin.addTerm(1.0, editVar);
            }
        }
        //set minimize function
        model.setObjective(exprMin, GRB.MINIMIZE);

        int heuValue = -1;
        if (useHeuristicUpperBound) {
            BlockPartitionResult heuRes = Heuristic.computeUpperBoundWithBlockPartition(g, t);
            model.addConstr(exprMin, GRB.LESS_EQUAL, heuRes.getCost(), "");
            heuValue = heuRes.getCost();
            setStartAssignmentAsHeuristic(heuRes);
        }
        log("heu res: " + heuValue);
    }

    private static void setStartAssignmentAsHeuristic(BlockPartitionResult heuRes) throws GRBException {
        BlockPartition blockPartition = heuRes.getBlockPartition();
        List<Block> blockList = blockPartition.getBlocks();
        for (int i = 0; i < blockList.size(); i++) {
            Block block1 = blockList.get(i);

            for (int j = i; j < blockList.size(); j++) {
                Block block2 = blockList.get(j);

                int edgesBetweenBlocks = blockPartition.getEdgeCountBetween(block1, block2);
                int maxEdges = (i == j) ? over2(block1.size) : block1.size * block2.size;
                boolean connected = edgesBetweenBlocks >= (maxEdges + 1) / 2;

                for (int uIndex = 0; uIndex < block1.size; uIndex++) {
                    Vertex u = block1.get(uIndex);
                    for (int vIndex = (i == j) ? uIndex + 1 : 0; vIndex < block2.size; vIndex++) {
                        Vertex v = block2.get(vIndex);
                        //set edit-variables
                        editArr[u.id][v.id].set(GRB.DoubleAttr.Start, (g.getEdge(u, v).isEdge != connected) ? 1.0 : 0.0);
                    }
                }
            }
        }
    }

    static GRBCallback callback = new GRBCallback() {
        @Override
        protected void callback() {
            if (where == GRB.Callback.MIPSOL) {
                mipSolCallbacks++;
                try {
                    transformGraphToCurrentSolution(g);

                    if (useMultipleConstraints) {
                        List<GRBLinExpr> exprs = createMultipleConstraints();
                        if (exprs != null)
                            for (GRBLinExpr expr : exprs) addLazy(expr, GRB.GREATER_EQUAL, 1.0);
                    } else {
                        GRBLinExpr expr = createConstraint();
                        if (expr != null) addLazy(expr, GRB.GREATER_EQUAL, 1.0);
                    }

                } catch (GRBException e) {
                    e.printStackTrace();
                }
            }
        }

        private void transformGraphToCurrentSolution(Graph graph) throws GRBException {
            for (int i = 0; i < graph.vertexCount; i++) {
                Vertex u = graph.get(i);
                for (int j = i + 1; j < graph.vertexCount; j++) {
                    Vertex v = graph.get(j);
                    Edge e = graph.getEdge(u, v);
                    GRBVar editVar = editArr[u.id][v.id];
                    int edited = (int) Math.round(getSolution(editVar));
                    boolean isCurrentEdge = Math.abs(edited - initEdgeArr[u.id][v.id]) == 1;
                    if (isCurrentEdge != e.isEdge) graph.editEdge(e);
                }
            }
        }
    };

    private static GRBLinExpr createConstraint() {
        List<Vertex> branchVertices = getBranchVertices();
        if (branchVertices.size() <= t) return null;

        return getLinExprOf(branchVertices);
    }

    private static LinkedList<GRBLinExpr> createMultipleConstraints() {
        var ncs = computeNCs();
        if (ncs.size() <= t) return null;
        LinkedList<GRBLinExpr> res = new LinkedList<>();

        for (int start = 0; start < ncs.size() - t; start++) {
            List<List<Vertex>> branchNcs = new LinkedList<>();
            for (int i = start; i < start + t + 1; i++) {
                branchNcs.add(ncs.get(i % ncs.size()));
            }
            createConstrains(branchNcs, res);
            if (useLimitForMultipleConstraints && res.size() == limit) break;
        }
        return res;
    }

    private static void createConstrains(List<List<Vertex>> branchNcs, LinkedList<GRBLinExpr> res) {
        LinkedList<Vertex> branchedVertices = new LinkedList<>();
        createConstrainsAux(0, branchedVertices, branchNcs, res);
    }

    private static void createConstrainsAux(int index, LinkedList<Vertex> branchedVertices, List<List<Vertex>> branchNCs, LinkedList<GRBLinExpr> res) {
        if (index == t + 1) {
            res.add(getLinExprOf(branchedVertices));
            return;
        }
        for (Vertex vertex : branchNCs.get(index)) {
            branchedVertices.add(vertex);
            createConstrainsAux(index + 1, branchedVertices, branchNCs, res);
            branchedVertices.removeLast();
        }
    }

    private static List<List<Vertex>> computeNCs() {
        List<List<Vertex>> ncs = new ArrayList<>();
        for (int i = 0; i < g.vertexCount; i++) {
            Vertex vertex = g.get(i);
            boolean newNCBoolean = true;
            for (List<Vertex> nc : ncs) {
                Vertex ncVertex = nc.get(0);
                Edge edge = g.getEdge(vertex, ncVertex);
                if (edge.getNumberOfNonCommons() == 0) {
                    newNCBoolean = false;
                    nc.add(vertex);
                    break;
                }
            }
            if (newNCBoolean) {
                List<Vertex> newNC = new LinkedList<>();
                newNC.add(vertex);
                ncs.add(newNC);
            }
        }
        return ncs;
    }

    private static GRBLinExpr getLinExprOf(List<Vertex> branchVertices) {
        List<Edge> unblockList = new LinkedList<>();

        GRBLinExpr expr = new GRBLinExpr();
        int editedCounter = 0;
        for (int i = 0; i < branchVertices.size(); i++) {
            Vertex u = branchVertices.get(i);
            for (int j = i + 1; j < branchVertices.size(); j++) {
                Vertex v = branchVertices.get(j);
                Edge branchEdge = g.getEdge(u, v);
                Vertex nonCommon = branchEdge.nonCommons.getLastElement();
                Edge uNonCommon = g.getEdge(u, nonCommon);
                Edge vNonCommon = g.getEdge(v, nonCommon);

                if (!uNonCommon.isBlocked) {
                    uNonCommon.block();
                    unblockList.add(uNonCommon);

                    if (uNonCommon.isEdited) { // ... + (1 - e1) + ... + (1 - e1) + e2 + e2 >= 1
                        editedCounter++;
                        expr.addTerm(-1.0, editArr[u.id][nonCommon.id]);
                    } else
                        expr.addTerm(1.0, editArr[u.id][nonCommon.id]);
                }
                if (!vNonCommon.isBlocked) {
                    vNonCommon.block();
                    unblockList.add(vNonCommon);

                    if (vNonCommon.isEdited) {
                        editedCounter++;
                        expr.addTerm(-1.0, editArr[v.id][nonCommon.id]);
                    } else
                        expr.addTerm(1.0, editArr[v.id][nonCommon.id]);
                }
            }
        }
        for (Edge edge : unblockList) edge.unblock();

        expr.addConstant(editedCounter);
        return expr;
    }

    public static List<Vertex> getBranchVertices() {
        List<Vertex> branchedVertices = new ArrayList<>();
        branchedVertices.add(g.get(0));

        for (int i = 1; i < g.vertexCount; i++) {
            if (branchedVertices.size() == t + 1) break;
            Vertex u = g.get(i);
            boolean newNC = true;
            for (Vertex v : branchedVertices) {
                Edge e = g.getEdge(u, v);
                if (e.getNumberOfNonCommons() == 0) {
                    newNC = false;
                    break;
                }
            }
            if (newNC) branchedVertices.add(u);
        }
        return branchedVertices;
    }

    private static int countEdits(Graph graph) throws GRBException {
        double sum = 0;
        for (int i = 0; i < graph.vertexCount; i++) {
            for (int j = i + 1; j < graph.vertexCount; j++) {
                GRBVar v = editArr[i][j];
                sum += v.get(GRB.DoubleAttr.X);
            }
        }
        int res = (int) Math.round(sum);
        log("k: " + res);
        return res;
    }

}

