package algo.ilp;

import algo.datastructure.Graph;
import algo.datastructure.Vertex;
import algo.heuristic.Heuristic;
import algo.util.Util;
import gurobi.*;

import java.io.*;

import static algo.util.Util.log;

public class MBILP {
    private static Graph g;
    private static int t;
    private static GRBVar[][] xArr;
    private static GRBVar[][][][] yArr;

    //set these variables to give Gurobi a time-limit in seconds
    public static boolean useTimeLimit = false;
    public static int timeLimit;

    public static boolean useHeuristicUpperBound = true;

    public static int solve(Graph graph, int t) throws IOException {
        long startTime = System.currentTimeMillis();
        MBILP.g = graph;
        log("t: " + t);
        int k;
        if (useHeuristicUpperBound)
            k = Heuristic.computeUpperBound(g, t);
        else
            k = Integer.MAX_VALUE;

        double accumulatedTime = 0;
        for (int currentT = 1; currentT <= t; currentT++) {
            MBILP.t = currentT;
            int bestMatrix = 0;
            try {
                GRBEnv env = new GRBEnv(true);
                env.set(GRB.IntParam.Threads, 1);
                env.set(GRB.IntParam.LazyConstraints, 1);
                if (!Util.printLogs)
                    env.set(GRB.IntParam.OutputFlag, 0);//comment if you want more information form Gurobi
                env.start();
                GRBModel model = new GRBModel(env);

                initializeVariables(graph, model);
                setConstraints(graph, model);
                //TODO add path of files here
                String fileName = "./src/algo/ilp/expfiles/non_iso_mats_" + currentT + ".txt";

                BufferedReader bi = new BufferedReader(new FileReader(fileName));

                int[][] imageMatrix = new int[currentT][currentT];
                int matrixCtr = 0;
                while (getNextMatrix(bi, currentT, imageMatrix)) {
                    System.out.println("start matrix t " + currentT + " matrix " + matrixCtr);
                    if (useTimeLimit && accumulatedTime > timeLimit) break;
                    matrixCtr++;
                    GRBLinExpr obj = setObjectiveFunction(graph, model, imageMatrix, currentT);
                    if (useTimeLimit) model.set(GRB.DoubleParam.TimeLimit, timeLimit - accumulatedTime);

                    GRBConstr constr = null;
                    if (k != Integer.MAX_VALUE)
                        constr = model.addConstr(obj, GRB.LESS_EQUAL, k - 1, "obj"); // small improvement: use best value as upper bound
                    model.optimize();
                    if (k != Integer.MAX_VALUE) model.remove(constr);

                    accumulatedTime += model.get(GRB.DoubleAttr.Runtime);

                    int status = model.get(GRB.IntAttr.Status);
                    if (status != GRB.INFEASIBLE && status != GRB.TIME_LIMIT) {
                        int val = (int) Math.round(model.get(GRB.DoubleAttr.ObjVal));
                        if (val < k) k = val;
                        bestMatrix = matrixCtr;
                    }
                }
                //dispose
                if (useTimeLimit && accumulatedTime > timeLimit) k = -1;//timeout
                model.dispose();
                env.dispose();
            } catch (GRBException | FileNotFoundException e) {
                e.printStackTrace();
                return -1;
            }
            System.out.println("k-t" + currentT + ": " + k);
            System.out.println("time-t" + currentT + ": " + (System.currentTimeMillis() - startTime));
            System.out.println("matrix-t" + currentT + ": " + bestMatrix);
        }
        System.out.println("k: " + k);
        return k;
    }

    private static boolean getNextMatrix(BufferedReader bi, int size, int[][] imageMatrix) throws IOException {
        String line;
        int rowCtr = 0;
        while ((line = bi.readLine()) != null && rowCtr < size) {
            if (line.startsWith("#")) continue;
            String[] row = line.split(" ");
            for (int i = 0; i < row.length; i++)
                imageMatrix[rowCtr][i] = Integer.parseInt(row[i]);
            rowCtr++;
        }
        return size == rowCtr;
    }

    private static GRBLinExpr setObjectiveFunction(Graph graph, GRBModel model, int[][] imageMatrix, int size) throws GRBException {
        GRBLinExpr obj = new GRBLinExpr();
        for (int p = 0; p < size; p++) {
            for (int l = 0; l < size; l++) {
                for (int i = 0; i < graph.vertexCount; i++) {
                    Vertex u = g.get(i);
                    for (int j = i + 1; j < graph.vertexCount; j++) { //note j starts at i + 1 since we consider only undirected graphs
                        Vertex v = g.get(j);
                        int b = imageMatrix[p][l];
                        int s = g.areNeighbors(u, v) ? 1 : 0;
                        int factor = b + s - 2 * b * s;
                        if (factor != 0)
                            obj.addTerm(factor, yArr[i][j][p][l]);
                    }
                }
            }
        }
        model.setObjective(obj, GRB.MINIMIZE);
        return obj;
    }


    private static void initializeVariables(Graph graph, GRBModel model) throws GRBException {
        int vertexCount = graph.vertexCount;
        //creating variables
        xArr = new GRBVar[vertexCount][t];
        yArr = new GRBVar[vertexCount][vertexCount][t][t];

        for (int i = 0; i < vertexCount; i++) {
            Vertex u = g.get(i);
            for (int p = 0; p < t; p++) {
                GRBVar xVar = model.addVar(0.0, 1.0, 0, GRB.BINARY, "x_" + u.id + "_" + p);
                xArr[u.id][p] = xVar;

                for (int j = 0; j < vertexCount; j++) {
                    Vertex v = g.get(j);
                    for (int l = 0; l < t; l++) {
                        GRBVar yVar = model.addVar(0.0, 1.0, 0, GRB.BINARY, "y_" + u.id + "_" + p + "_" + v.id + "_" + l);
                        yArr[u.id][v.id][p][l] = yVar;
                    }
                }
            }
        }
    }

    private static void setConstraints(Graph graph, GRBModel model) throws GRBException {
        // line (b) -> see orignal paper
        for (int i = 0; i < graph.vertexCount; i++) {
            Vertex u = graph.get(i);
            int uIndex = u.id;
            GRBLinExpr expr = new GRBLinExpr();
            for (int p = 0; p < t; p++) {
                expr.addTerm(1.0, xArr[uIndex][p]);
            }
            model.addConstr(expr, GRB.EQUAL, 1, "");
        }
        // line (c) -> see orignal paper
        for (int p = 0; p < t; p++) {
            GRBLinExpr expr = new GRBLinExpr();
            for (int i = 0; i < graph.vertexCount; i++) {
                expr.addTerm(1.0, xArr[graph.get(i).id][p]);
            }
            model.addConstr(expr, GRB.GREATER_EQUAL, 1, "");

        }
        // yArr
        for (int i = 0; i < graph.vertexCount; i++) {
            Vertex u = graph.get(i);
            int uIndex = u.id;
            for (int j = 0; j < graph.vertexCount; j++) {
                Vertex v = graph.get(j);
                int vIndex = v.id;
                for (int p = 0; p < t; p++) {
                    for (int l = 0; l < t; l++) {
                        GRBVar y = yArr[uIndex][vIndex][p][l];
                        GRBVar xi = xArr[i][p];
                        GRBVar xj = xArr[j][l];

                        model.addConstr(y, GRB.LESS_EQUAL, xi, "");
                        model.addConstr(y, GRB.LESS_EQUAL, xj, "");
                        GRBLinExpr expr = new GRBLinExpr();
                        expr.addTerm(1, xi);
                        expr.addTerm(1, xj);
                        expr.addConstant(-1);
                        model.addConstr(y, GRB.GREATER_EQUAL, expr, "");
                        model.addConstr(y, GRB.GREATER_EQUAL, 0, "");
                    }
                }
            }
        }
    }
}
